package RDF::SPARUL::Redland;

BEGIN {
	$RDF::SPARUL::Redland::VERSION = '0.01';
}

=head1 NAME

RDF::SPARUL::Redland - SPARUL (SPARQL Update) support for RDF::Redland.

=cut

=head1 VERSION

0.01

=cut

use 5.010000;
use strict;

use Carp;
use LWP::UserAgent;
use RDF::Redland;
use Text::Balanced qw(extract_bracketed extract_delimited);
use URI::URL;

require Exporter;
our @ISA = qw(Exporter);

our %EXPORT_TAGS = ( 'all' => [ qw(
	sparul
) ] );

our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

our @EXPORT = qw(
	
);

=head1 SYNOPSIS

  use RDF::Redland;
  use RDF::SPARUL::Redland;
  
  my $script = <<SCRIPTEND;
  PREFIX rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
  PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
  PREFIX dc:   <http://purl.org/dc/terms/>
  LOAD <http://example.com/some.rdf>
  INSERT { ?thing rdfs:comment "This has a title." }
     WHERE { ?thing dc:title ?title }
  SCRIPTEND
  
  my $storage = RDF::Redland::Storage->new("hashes", "test", "new='yes',hash-type='memory'");
  my $model   = RDF::Redland::Model->new($storage, "");
  
  my $sparul  = RDF::SPARUL::Redland->($model);
  $sparul->prepare($script);
  $sparul->execute;
  
  print $model->as_string;

=head1 DESCRIPTION

SPARQL is an SQL-like query language for RDF. However, it is read-only. That is,
it cannot make any changes to the RDF database. SPARUL (SPARQL Update) is an
extension to RDF to allow modifications to the RDF database.

This module wraps around an RDF::Redland::Model object to provide support for
SPARUL. It partly works.

=head1 FUNCTIONAL INTERFACE

For those not wanting to use the object-oriented interface, a simple function
is provided.

=over 8

=item sparul($model, $script)

Runs the script on the model. Returns nothing in particular.

=back

=cut

sub sparul
{
	my $model  = shift;
	my $script = shift;
	
	my $sparul = RDF::SPARUL::Redland->new($model);
	$sparul->prepare($script)->execute;
}

=head1 OBJECT-ORIENTED INTERFACE

=head2 Constructor

=over 8

=item $sparul = RDF::SPARUL::Redland->new($model)

Creates a new SPARUL object. The model passed to the constructor will become the
default graph used by SPARUL scripts.

If $model is omitted, a blank RDF::Redland::Model is created.

=back

=cut
                    
sub new
{
	my $class  = shift;
	my $model  = shift;
	
	unless (defined $model)
	{
		my $storage = RDF::Redland::Storage->new("hashes", "test", "new='yes',hash-type='memory'");
		$model   = RDF::Redland::Model->new($storage, "");
	}
	
	my $self = bless { 'models' => { '_' => $model } }, $class;
	return $self;
}

=head2 Public Methods

=over 8

=item $sparul->add_graph($model, $uri);

Adds an additional model to the SPARUL object with a given URI. SPARUL supports
multiple named graphs.

=cut

sub add_graph
{
	my $self  = shift;
	my $model = shift;
	my $uri   = shift;
	
	$self->{'models'}->{ $uri } = $model;
}

=item $model = $sparul->graph($uri);

Get an RDF::Redland::Model representing a named graph. This might be a model
previously added using C<add_graph>, or a model created by the script itself
using C<CREATE GRAPH>.

Of the $uri is omitted, returns the default graph.

=cut

sub graph
{
	my $self = shift;
	my $graph_uri = shift;
	
	my $graph = $self->{'models'}->{'_'};
	$graph = $self->{'models'}->{ $self->absolute($graph_uri) }
		if defined $graph_uri;
		
	return $graph;
}

=item $sparul->prepare($script);

Parse a SPARUL script, ready for execution.

=cut

sub prepare
{
	my $self   = shift;
	my $script = shift;

	delete $self->{'base'};
	delete $self->{'prefixes'};
	delete $self->{'templates'};
	delete $self->{'statements'};
	
	$self->set_script($script);
	$self->{'statements'} = [ _Script($self) ];
	return $self;
}

=item $sparul->execute;

Execute the previously prepared script. This cannot be usefully called before
C<prepare>.

=cut

sub execute
{
	my $self = shift;

	return 0 unless $self->{'statements'};

	STATEMENT: foreach my $q (@{ $self->{'statements'} })
	{
		# These don't modify the data, so we ignore them.
		next if $q->{'type'} =~ /^(ASK|CONSTRUCT|DESCRIBE|SELECT)$/i;

		my $graph = $self->graph($q->{'graph'});

		if ($q->{'type'} =~ /^(MODIFY|DELETE|INSERT)/)
		{
			last STATEMENT
				unless $self->_modify($graph, $q->{'conditions'}, $q->{'template_delete'}, $q->{'template_insert'});
		}
		
		elsif ($q->{'type'} =~ /^(CLEAR)$/)
		{
			last STATEMENT
				unless $self->_modify($graph, '{?s ?p ?o}', '{?s ?p ?o}', undef);
		}
		
		elsif ($q->{'type'} =~ /^(CREATE)$/)
		{
			my $graph_uri = $self->absolute($q->{'graph'});
			if (defined $self->{'models'}->{ $graph_uri })
				{ last STATEMENT unless length $q->{'silent'}; }
			else
			{
				my $storage = RDF::Redland::Storage->new("hashes", "test", "new='yes',hash-type='memory'");
				my $model   = RDF::Redland::Model->new($storage, "");
				$self->{'models'}->{ $graph_uri } = $model;
			}
		}

		elsif ($q->{'type'} =~ /^(DROP)$/)
		{
			my $graph_uri = $self->absolute($q->{'graph'});
			if (defined $self->{'models'}->{ $graph_uri })
				{ $self->{'models'}->{ $graph_uri } = undef; }
			else
				{ last STATEMENT unless length $q->{'silent'}; }
		}

		elsif ($q->{'type'} =~ /^(LOAD)$/)
		{
			$graph->load( RDF::Redland::URI->new($self->absolute($q->{'source'})) );
		}
	}
	return $self;
}

# for internal use only
sub script
{
	my $self = shift;
	return $self->{'script'};
}

# for internal use only
sub clean
{
	my $self = shift;
	$self->{'script'} =~ s/^[\s\r\n]+//gs;
}

# for internal use only
sub set_script
{
	my $self   = shift;
	my $script = shift;

	$self->{'script'} = $script;
	$self->clean;
}

# for internal use only
sub eat
{
	my $self = shift;
	my $re   = shift;
	
	if ($self->script =~ /^$re/)
	{
		$self->set_script( substr($self->script, length $&) );
	
		return [$&, $1, $2, $3, $4, $5, $6, $7, $8, $9];
	}
	
	confess "Was expecting something that matched $re but got $self->{script}\n";
}

# for internal use only
sub optional_terminator
{
	my $self = shift;
	$self->eat(qr/(\.|\;)/);
}

# for internal use only
sub eye
{
	my $self = shift;
	my $re   = shift;
	
	if ($self->script =~ /^$re/)
	{
		return [$&, $1, $2, $3, $4, $5, $6, $7, $8, $9];
	}
	
	return 0;
}

# for internal use only
sub done
{
	my $self = shift;
	$self->clean;
	return ($self->script eq '');
}

# for internal use only
sub absolute
{
	my $self  = shift;
	my $param = shift;

	if ($param =~ /^([a-z][a-z0-9\+\.\-]*)\:/i)
	{
		# seems to be an absolute URI, so can safely return "as is".
		return $param;
	}

	my $url = url $param, $self->{'base'};
	return $url->abs->as_string;
}

# for internal use only
sub _modify
{
	my $self   = shift;
	my $model  = shift;
	my $cond   = shift;
	my $remove = shift;
	my $add    = shift;
	
	# Keep backup of $model.
	my $backup = RDF::Redland::Model->new_from_model($model);

	my @RS;

	if ($cond)
	{
		my $qstr;	
		foreach my $p (keys %{ $self->{'prefixes'} })
		{
			$qstr .= sprintf("PREFIX %s: <%s>\n", $p, $self->{'prefixes'}->{$p});
		}
		$qstr .= "SELECT * WHERE $cond\n";
		
		my $query = RDF::Redland::Query->new($qstr, undef, undef, "sparql");
		my $results = $model->query_execute($query);
		while (!$results->finished)
		{
			my $r = {};
			for (my $i=0; $i < $results->bindings_count(); $i++)
			{
				my $key = $results->binding_name($i);
				my $val = $results->binding_value($i);
				$r->{$key} = $val;
			}
			push @RS, $r;
			$results->next_result;
		}
	}
	else
	{
		@RS = ({});
	}
	
	$self->{'node'} = {};
	
	if ($remove)
	{
		foreach my $r (@RS)
		{
			my @triples_for_removal = $self->template_substitute($remove, $r);
			foreach my $triple (@triples_for_removal)
				{ $model->remove_statement(@$triple); }
		}
	}
	
	if ($add)
	{
		foreach my $r (@RS)
		{
			my @triples_for_addition = $self->template_substitute($add, $r);
			foreach my $triple (@triples_for_addition)
			{
				$model->add_statement($triple->[0], $triple->[1], $triple->[2]);
			}
		}
	}
	
	$self->{'node'} = {};

	return 1;
}

# for internal use only
sub template_tokenise
{
	my $self = shift;
	my $ts   = shift;
	my @rv;
	
	if ($ts =~ /^[\s\r\n]*\{.*\}[\s\r\n]*$/s)
	{
		$ts =~ s/^([\s\r\n]*\{)|(\}[\s\r\n]*$)//gs;
	}
	
	while (length $ts)
	{
		$ts =~ s/^[\s\r\n]+//s;
		
		if ($ts =~ /^([\.\,\;])/)
		{
			push @rv, $1;
			$ts = substr($ts, length $1);
		}
		elsif ($ts =~ /^([\?\$]\S+)/)
		{
			push @rv, $1;
			$ts = substr($ts, length $1);
		}
		elsif ($ts =~ /^(<[^>]*>)/)
		{
			push @rv, $1;
			$ts = substr($ts, length $1);
		}
		elsif ($ts =~ /^"/)
		{
			my $string;
			($string, $ts) = extract_delimited($ts, '"');
			
			if ($ts =~ /^(\@[^\^\s\;\,\.\]\[]+)/)
			{
				$string .= $1;
				$ts = substr($ts, length $1);
			}
			if ($ts =~ /^(\^\^<[^>]*>)/)
			{
				$string .= $1;
				$ts = substr($ts, length $1);
			}
			elsif ($ts =~ /^(\^\^[^:]*:[^\s\,\;]*)/)
			{
				$string .= $1;
				$ts = substr($ts, length $1);
			}
			
			push @rv, $string;
		}
		elsif ($ts =~ /^(_:[^\s\,\;]*)/)
		{
			push @rv, $1;
			$ts = substr($ts, length $1);
		}
		elsif ($ts =~ /^([^\s\,\;]*:[^\s\,\;]*)/)
		{
			push @rv, $1;
			$ts = substr($ts, length $1);
		}
	}
	
	return @rv;
}

# for internal use only
sub template_parse
{
	my $self            = shift;
	my $template_string = shift;
	
	my @tokens = $self->template_tokenise($template_string);
	my $triples;
	my $current_triple;
	
	while (my $t = shift @tokens)
	{
		if ($t eq '.')
		{
			push @$triples, $current_triple;
			$current_triple = [];
		}
		elsif ($t eq ';')
		{
			push @$triples, $current_triple;
			$current_triple = [$current_triple->[0]];
		}
		elsif ($t eq ',')
		{
			push @$triples, $current_triple;
			$current_triple = [$current_triple->[0] , $current_triple->[1]];
		}
		else
		{
			push @$current_triple, $t;
		}
	}
	
	if ($current_triple->[2])
		{ push @$triples, $current_triple; }
	
	return $triples;
}

# for internal use only
sub template_substitute
{
	my $self            = shift;
	my $template_string = shift;
	my $substitutions   = shift;
	
	unless (defined $self->{'templates'}->{$template_string})
	{
		$self->{'templates'}->{$template_string} = $self->template_parse($template_string);
	}
	
	my @rv;
	foreach my $triple (@{ $self->{'templates'}->{$template_string} })
	{
		my @rv_triple;
		foreach my $node (@$triple)
		{
			if ($node =~ /^[\?\$](.+)$/)
			{
				push @rv_triple, $substitutions->{$1};
			}
			elsif ($node =~ /^(_:(.+))$/)
			{
				$self->{'node'}->{$1} = RDF::Redland::BlankNode->new($2)
					unless (defined $self->{'node'}->{$1});
				push @rv_triple, $self->{'node'}->{$1};
			}
			elsif ($node =~ /^<(.*)>$/)
			{
				$self->{'node'}->{$1} = RDF::Redland::URINode->new($1)
					unless (defined $self->{'node'}->{$1});
				push @rv_triple, $self->{'node'}->{$1};
			}
			elsif ($node =~ /^"/)
			{
				unless ($self->{'node'}->{$node})
				{
					my ($text, $remainder) = extract_delimited($node, '"');
					$text =~ s/(^\"|\"$)//sg;
					
					my ($lang, $dt);
					if ($remainder =~ /^\@([^\^\s]+)/)
						{ $lang = $1; }
					if ($remainder =~ /^\^\^<([.*])>/)
						{ $dt = $1; }
					elsif ($remainder =~ /^\^\^([.*]):([.*])/)
						{ $dt = $self->{'prefixes'}->{$1} . $2; }
						
					$self->{'node'}->{$node} = RDF::Redland::LiteralNode->new($text, $dt, $lang);
				}
				push @rv_triple, $self->{'node'}->{$node};
			}
			elsif ($node =~ /^(.*):(.*)$/)
			{
				my $uri = $self->{'prefixes'}->{$1} . $2;
				$self->{'node'}->{$uri} = RDF::Redland::URINode->new($uri)
					unless (defined $self->{'node'}->{$uri});
				push @rv_triple, $self->{'node'}->{$uri};
			}
		}
		push @rv, \@rv_triple
			if (defined $rv_triple[2]);
	}
	
	return @rv;
}


# for internal use only
sub _Script
{
	my $self = shift;
	my @rv;

	while ($self->eye(qr/(\.\;)/))
		{ $self->optional_terminator; }
	
	$self->{'base'}     = $self->_BaseDeclaration;
	$self->{'prefixes'} = $self->_PrefixDeclarations;

	while (!$self->done)
	{
		push @rv, $self->_Statement;

		while ($self->eye(qr/(\.\;)/))
			{ $self->optional_terminator; }
	}
	
	return @rv;
}

# for internal use only
sub _BaseDeclaration
{
	my $self = shift;

	if ($self->eye(qr/BASE\s*<([^>]*)>/i))
	{
		return $self->eat(qr/BASE\s*<([^>]*)>/i)->[1];
	}

	while ($self->eye(qr/(\.\;)/))
		{ $self->optional_terminator; }
}

# for internal use only
sub _PrefixDeclarations
{
	my $self = shift;
	my $rv   = {};

	while ($self->eye(qr/PREFIX/i))
	{
		my $r   = $self->eat(qr/PREFIX\s*(\S*)\s*:\s*<([^>]*)>/i);
		my $pfx = $r->[1];
		my $uri = $r->[2];
		
		$pfx = -1 unless length $pfx;
		$rv->{$pfx} = $uri;

		while ($self->eye(qr/(\.\;)/))
			{ $self->optional_terminator; }
	}
	
	return $rv;
}

# for internal use only
sub _Statement
{
	my $self = shift;
	
	if ($self->eye(qr/INSERT/i))
		{ return $self->_InsertStatement; }
		
	if ($self->eye(qr/DELETE/i))
		{ return $self->_DeleteStatement; }

	if ($self->eye(qr/CLEAR/i))
		{ return $self->_ClearStatement; }

	if ($self->eye(qr/LOAD/i))
		{ return $self->_LoadStatement; }

	if ($self->eye(qr/CREATE/i))
		{ return $self->_CreateStatement; }

	if ($self->eye(qr/DROP/i))
		{ return $self->_DropStatement; }

	if ($self->eye(qr/MODIFY/i))
		{ return $self->_ModifyStatement; }

	if ($self->eye(qr/SELECT/i))
		{ return $self->_SelectStatement; }

	if ($self->eye(qr/ASK/i))
		{ return $self->_AskStatement; }

	if ($self->eye(qr/DESCRIBE/i))
		{ return $self->_DescribeStatement; }

	if ($self->eye(qr/CONSTRUCT/i))
		{ return $self->_ConstructStatement; }
		
	confess "Unknown type of statement with $self->{'script'}.\n";
}

# for internal use only
sub _InsertStatement
{
	my $self = shift;
	my $rv = {};
	
	$rv->{'type'} = uc($self->eat(qr/INSERT/i)->[0]);
	$rv->{'data'} = $self->eat(qr/(DATA)?/i)->[0];
	
	$rv->{'graph'} = $self->eat(qr/INTO \s* <([^>]*)>/ix)->[1]
		if $self->eye(qr/INTO/i);

	my ($template, $rest) = extract_bracketed($self->script, '{"}');
	$rv->{'template_insert'} = $template;
	$self->set_script($rest);
	
	if ($self->eye(qr/WHERE/i))
	{
		$self->eat(qr/WHERE/i);
		
		my ($cond, $rest) = extract_bracketed($self->script, '{"}');
		$rv->{'conditions'} = $cond;
		$self->set_script($rest);
	}
	
	return $rv;
}

# for internal use only
sub _DeleteStatement
{
	my $self = shift;
	my $rv = {};
	
	$rv->{'type'} = uc($self->eat(qr/DELETE/i)->[0]);
	$rv->{'data'} = $self->eat(qr/(DATA)?/i)->[0];
	$rv->{'graph'} = $self->eat(qr/FROM \s* <([^>]*)>/ix)->[1]
		if $self->eye(qr/FROM/i);
	
	my ($template, $rest) = extract_bracketed($self->script, '{"}');
	$rv->{'template_delete'} = $template;
	$self->set_script($rest);
	
	if ($self->eye(qr/WHERE/i))
	{
		$self->eat(qr/WHERE/i);
		
		my ($cond, $rest) = extract_bracketed($self->script, '{"}');
		$rv->{'conditions'} = $cond;
		$self->set_script($rest);
	}
	
	return $rv;
}

# for internal use only
sub _ClearStatement
{
	my $self = shift;
	my $rv = {};
	
	$rv->{'type'} = uc($self->eat(qr/CLEAR/i)->[0]);
	$rv->{'graph'} = $self->eat(qr/GRAPH \s* <([^>]*)>/ix)->[1]
		if $self->eye(qr/GRAPH/i);
	
	return $rv;
}

# for internal use only
sub _LoadStatement
{
	my $self = shift;
	my $rv = {};
	
	$rv->{'type'} = uc($self->eat(qr/LOAD/i)->[0]);
	$rv->{'source'} = $self->eat(qr/<([^>]*)>/i)->[1];
	$rv->{'graph'} = $self->eat(qr/INTO \s* <([^>]*)>/ix)->[1]
		if $self->eye(qr/INTO/i);
	
	return $rv;
}

# for internal use only
sub _CreateStatement
{
	my $self = shift;
	my $rv = {};
	
	$rv->{'type'} = uc($self->eat(qr/CREATE/i)->[0]);
	$rv->{'silent'} = $self->eat(qr/(SILENT)?/i)->[0];
	$rv->{'graph'} = $self->eat(qr/GRAPH \s* <([^>]*)>/ix)->[1]
		if $self->eye(qr/GRAPH/i);
	
	return $rv;
}

# for internal use only
sub _DropStatement
{
	my $self = shift;
	my $rv = {};
	
	$rv->{'type'} = uc($self->eat(qr/DROP/i)->[0]);
	$rv->{'silent'} = $self->eat(qr/(SILENT)?/i)->[0];
	$rv->{'graph'} = $self->eat(qr/GRAPH \s* <([^>]*)>/ix)->[1]
		if $self->eye(qr/GRAPH/i);
	
	return $rv;
}

# for internal use only
sub _ModifyStatement
{
	my $self = shift;
	my $rv = {};
	
	$rv->{'type'} = uc($self->eat(qr/MODIFY/i)->[0]);
	$rv->{'graph'} = $self->eat(qr/((GRAPH\s*)?<([^>*])>)?/i)->[3];
	
	$self->eat(qr/DELETE/i);
	my ($template, $rest) = extract_bracketed($self->script, '{"}');
	$rv->{'template_delete'} = $template;
	$self->set_script($rest);

	$self->eat(qr/INSERT/i);
	($template, $rest) = extract_bracketed($self->script, '{"}');
	$rv->{'template_insert'} = $template;
	$self->set_script($rest);
	
	if ($self->eye(qr/WHERE/i))
	{
		$self->eat(qr/WHERE/i);
		
		my ($cond, $rest) = extract_bracketed($self->script, '{"}');
		$rv->{'conditions'} = $cond;
		$self->set_script($rest);
	}
	
	return $rv;
}

# for internal use only
sub _AskStatement
{
	my $self = shift;
	my $rv = {};
	
	$rv->{'type'} = uc($self->eat(qr/ASK/i)->[0]);
	
	while ($self->eye(qr/\{/i))
	{
		my ($clause, $rest) = extract_bracketed($self->script, '{"}');
		push @{ $rv->{'clauses'} }, $clause;
		$self->set_script($rest);
	}
	
	if ($self->eye(qr/WHERE/i))
	{
		$self->eat(qr/WHERE/i);
		
		my ($cond, $rest) = extract_bracketed($self->script, '{"}');
		$rv->{'conditions'} = $cond;
		$self->set_script($rest);
	}

	return $rv;
}

# for internal use only
sub _ConstructStatement
{
	my $self = shift;
	my $rv = {};
	
	$rv->{'type'} = uc($self->eat(qr/CONSTRUCT/i)->[0]);
	
	my ($clause, $rest) = extract_bracketed($self->script, '{"}');
	$rv->{'template_construct'} = $clause;
	$self->set_script($rest);
	
	# @@TODO: DatasetClause*

	if ($self->eye(qr/WHERE/i))
	{
		$self->eat(qr/WHERE/i);
		
		my ($cond, $rest) = extract_bracketed($self->script, '{"}');
		$rv->{'conditions'} = $cond;
		$self->set_script($rest);
	}

	# @@TODO: SolutionModifier

	return $rv;
}

# for internal use only
sub _DescribeStatement
{
	# @@TODO
	die "DESCRIBE not supported yet.\n";
}

# for internal use only
sub _SelectStatement
{
	# @@TODO
	die "SELECT not supported yet.\n";
}

1;

__END__

=back

=head1 BUGS

Many, many. Some major ones include:

=over 8

=item * Square brackets not supported in INSERT, DELETE or MODIFY templates.

=item * DESCRIBE, SELECT and CONSTRUCT not supported. These would essentially
be no-ops anyway. You should only include things in your SPARUL scripts which
would actually modify the data.

=item * WHERE patterns are limited by Redland's own limitations.

=item * Error handling is virtually non-existant. Syntax problems are apt to
cause infinite loops!

=back

=head1 SEE ALSO

L<RDF::Redland>

L<http://jena.hpl.hp.com/~afs/SPARQL-Update.html>

L<http://lists.perlrdf.org/listinfo/dev>

=head1 AUTHOR

Toby Inkster, E<lt>mail@tobyinkster.co.ukE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2009 by Toby Inkster

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.10.0 or,
at your option, any later version of Perl 5 you may have available.

=cut
